@@ -1,9 +1,12 @@
 #!/usr/bin/env python
+from dataclasses import dataclass
+from click import password_option
 from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
 from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives import hashes
 from cryptography.fernet import Fernet
 from threading import Thread
+import datetime
 import cryptography
 import asyncio
 import json
@@ -12,7 +15,6 @@ import time
 import sys
 import ssl
 import os
-import pickle
 
 if sys.platform == "win32":
     directory_cutter = "\\"
@@ -139,11 +141,11 @@ def prepData(data, metaData=None):
     # Prepares the data to be sent
     # Structure: DATA:| <data_length>.zfill(18) <raw-data> META:| <meta-string>
     # (ignore spaces)
-    data = pickle.dumps(data)
+    data = make_bytes(json.dumps(data))
     pData = ""
     pData = b"DATA:|" + str(len(data)).encode().zfill(18) + data
     if metaData:
-        pData = pData + b"META:|" + pickle.dumps(metaData)
+        pData = pData + b"META:|" + make_bytes(json.dumps(metaData))
     return pData
 
 
@@ -156,17 +158,67 @@ def dissectData(data):
         dataLen = int(data[:18])  # Extract length of data
         data = data[18:]  # Remove the data length
         rawData = data[:dataLen]  # Get the raw data
-        rawData = pickle.loads(rawData)  # Decode the data
+        rawData = json.loads(rawData)  # Decode the data
         metaStr = data[dataLen:]  # Get the meta-data (if any)
         if metaStr != "":
             if metaStr.startswith(b"META:|"):  # Received Meta-Data
                 metaStr = metaStr.lstrip(b"META:|")  # Remove "META:|"
-                metaData = pickle.loads(metaStr)
+                metaData = json.loads(metaStr)
     else:
         return data, None, True
     return rawData, metaData, False
 
 
+@dataclass
+class Login_Attempt():
+    timestamp: float
+    username: str
+    address: str
+    
+    def to_datetime(self):
+        return datetime.datetime.fromtimestamp(self.timestamp)
+    
+    def to_string(self):
+        return "{} - {} - {}".format(self.to_datetime(), self.username, self.address)
+    
+    def to_json(self):
+        return {
+            "timestamp": self.timestamp,
+            "username": self.username,
+            "address": self.address
+        }
+    
+    def from_json(self, data):
+        self.timestamp = data["timestamp"]
+        self.username = data["username"]
+        self.address = data["address"]
+        return self
+    
+    def __eq__(self, other):
+        return self.timestamp == other.timestamp and self.username == other.username and self.address == other.address
+    
+    def __hash__(self):
+        return hash(self.timestamp) ^ hash(self.username) ^ hash(self.address)
+    
+    def __lt__(self, other):
+        return self.timestamp < other.timestamp
+    
+    def __le__(self, other):
+        return self.timestamp <= other.timestamp
+    
+    def __gt__(self, other):
+        return self.timestamp > other.timestamp
+    
+    def __ge__(self, other):
+        return self.timestamp >= other.timestamp
+    
+    def __ne__(self, other):
+        return self.timestamp != other.timestamp or self.username != other.username or self.address != other.address
+    
+    def __str__(self):
+        return self.to_string()
+
+
 class User:
     def __init__(self, username):
         self.username = username
@@ -181,8 +233,6 @@ class User:
         #                         Login 1                        Login 2
 
         self.loginAttempts = []
-        # Structure: [ [<time.time()>, <IP_Address>], [<time.time()>, <IP_Address>] ]
-        #                        Attempt 1                      Attempt 2
 
     def encryptData(self, data):
         if self.hasPassword and self.password:
@@ -208,7 +258,7 @@ class User:
             "username": self.username,
             "hasPassword": self.hasPassword,
             "loginHistory": self.loginHistory,
-            "loginAttempts": self.loginAttempts,
+            "loginAttempts": [attempt.to_json() for attempt in self.loginAttempts],
         }
 
         if self._cipher_pass and self.hasPassword:
@@ -247,7 +297,7 @@ class User:
             self.username = user_info["username"]
             self.hasPassword = user_info["hasPassword"]
             self.loginHistory = user_info["loginHistory"]
-            self.loginAttempts = user_info["loginAttempts"]
+            self.loginAttempts = [Login_Attempt(**attempt) for attempt in user_info["loginAttempts"]]
             return self
 
     def verify(self, password):
@@ -273,7 +323,8 @@ class User:
             connection.verifiedUser = True
             connection.currentUser = self
             return True
-        self.loginAttempts.append([time.time(), connection.addr])
+        new_login_attempt = Login_Attempt(time.time(), username, connection.addr)
+        self.loginAttempts.append(new_login_attempt)
         return False
 
     def logout(self, client):
@@ -320,7 +371,7 @@ class Connection:
                 #       <current buffer length>, <target buffer length>
         return 0
 
-    async def send_data(self, data, metaData=None, enc=True):
+    async def _send_data(self, data, metaData=None, enc=True):
         data = prepData(data, metaData=metaData)
 
         if self.verifiedUser and enc and self._usr_enc:
@@ -335,7 +386,7 @@ class Connection:
         if self.server.loop:
             try:
                 asyncio.run_coroutine_threadsafe(
-                    self.send_data(data, metaData=metaData, enc=enc), self.server.loop
+                    self._send_data(data, metaData=metaData, enc=enc), self.server.loop
                 )
             except Exception as e:
                 print("Error sending data")
@@ -422,8 +473,6 @@ class Host:
         self._enable_buffer_monitor = True
 
         self.loginAttempts = []
-        # Structure: # Structure: [ [<time.time()>, <IP_Address>], [<time.time()>, <IP_Address>] ]
-        #                                     Attempt 1                      Attempt 2
 
         self.blacklistThreshold = 1800  # (In seconds)
         # If too many login attempts are made within this threshold, the address will be blacklisted
@@ -463,14 +512,17 @@ class Host:
         }
         self.useTermColors = useTermColors
 
-    def __pack_server_info__(self):
+    def _pack_server_info(self):
         server_info = {}
         for sVar in self._save_vars:
             if sVar in self.__dict__:
-                server_info[sVar] = self.__dict__[sVar]
+                if sVar == "loginAttempts":
+                    server_info[sVar] = [attempt.to_json() for attempt in self.loginAttempts]
+                else:
+                    server_info[sVar] = self.__dict__[sVar]
         return server_info
 
-    def __start_loop__(self, loop, task, finishFunc):
+    def _start_loop(self, loop, task, finishFunc):
         asyncio.set_event_loop(loop)
         loop.run_until_complete(asyncio.ensure_future(task()))
         if finishFunc:
@@ -478,18 +530,46 @@ class Host:
 
     def newLoop(self, task, finishFunc=None):
         new_loop = asyncio.new_event_loop()
-        t = Thread(target=self.__start_loop__, args=(new_loop, task, finishFunc))
+        t = Thread(target=self._start_loop, args=(new_loop, task, finishFunc))
         t.start()
 
+    def get_login_attempts(self, username=None, address=None, start_date=None, end_date=None):
+        attempts = self.loginAttempts[:]
+
+        # Change start_date and end_date to float timestamps if possible
+        if isinstance(start_date, datetime.datetime):
+            start_date = start_date.timestamp()
+        if isinstance(end_date, datetime.datetime):
+            end_date = end_date.timestamp()
+        
+        if username is not None:
+            # Filter out all attempts that do not match the username
+            attempts = [a for a in attempts if a.username == username]
+        
+        if address is not None:
+            # Filter out all attempts that do not match the address
+            attempts = [a for a in attempts if a.address == address]
+        
+        if start_date is not None:
+            # Filter out all login attempts before the start date
+            attempts = [a for a in attempts if a.timestamp >= start_date]
+        
+        if end_date is not None:
+            # Filter out all login attempts after the end date
+            attempts = [a for a in attempts if a.timestamp <= end_date]
+        
+        return self.loginAttempts
+
     def save(self, location, password=None):
         location = location.rstrip(directory_cutter)
         base_name = os.path.basename(location)
         location_directory = location.rstrip(base_name)
+        
         if os.path.exists(location_directory) or location.count(directory_cutter) == 0:
             if not (location.endswith(".json")):
                 location += ".json"
             with open(location, "w") as f:
-                json.dump(self.__pack_server_info__(), f)
+                json.dump(self._pack_server_info(), f)
             if password:
                 if isinstance(password, str):
                     password = password.encode()
@@ -503,16 +583,24 @@ class Host:
                         password = password.encode()
                     if isinstance(password, bytes):
                         decryptFile(location, password)
+                
                 with open(location, "rb") as f:
                     server_info = json.load(f)
+                
                 if password:
                     if isinstance(password, bytes):
                         encryptFile(location, password)
                 for sVar in self._save_vars:
                     if sVar in self.__dict__ and sVar in server_info:
-                        self.__dict__[sVar] = server_info[sVar]
+                        if sVar == "loginAttempts":
+                            self.loginAttempts = [Login_Attempt(**attempt) for attempt in server_info[sVar]]
+                        else:
+                            self.__dict__[sVar] = server_info[sVar]
 
-    def loadUsers(self):
+    def loadUsers(self, customPath=None):
+        if customPath is not None:
+            self.userPath = customPath
+        
         self.log("Loading users...")
         for i in os.listdir(self.userPath):
             iPath = os.path.join(self.userPath, i)
@@ -521,7 +609,10 @@ class Host:
                 self.users[user.username] = user
         self.log("Users loaded")
 
-    def saveUsers(self):
+    def saveUsers(self, customPath=None):
+        if customPath is not None:
+            self.userPath = customPath
+        
         self.log("Saving users...")
         for username in self.users:
             savePath = self.users[username].save(self.userPath)
@@ -542,7 +633,7 @@ class Host:
             return True
         return False
 
-    async def __add_to_log__(self, text, modifier, blinking):
+    async def _add_to_log(self, text, modifier, blinking):
         await self._lock.acquire()
 
         if modifier:
@@ -589,7 +680,7 @@ class Host:
         if not self.loop:
             print("Loop not running - unable to log text")
         asyncio.run_coroutine_threadsafe(
-            self.__add_to_log__(text, modifier, blinking), self.loop
+            self._add_to_log(text, modifier, blinking), self.loop
         )
 
     def gotData(self, client, data, metaData):
@@ -631,7 +722,7 @@ class Host:
         else:
             self.blacklisted(addr)
 
-    def __buffer_monitor__(self, client, reader):
+    def _buffer_monitor(self, client, reader):
         client.downloading = False
         while self.running and not client.writer.is_closing():
             if (
@@ -667,7 +758,7 @@ class Host:
             raise e
         return data.rstrip(self.sepChar)
 
-    async def __got_login_info__(self, client, username, password):
+    async def _got_login_info(self, client, username, password):
         self.log("Login acquired - verifying {}...".format(client.addr), "yellow")
         user = self.users[username]
         await asyncio.sleep(self.loginDelay)
@@ -688,14 +779,14 @@ class Host:
             ),
             "red_bg",
         )
-        self.loginAttempts.append([time.time(), client.addr])
+        new_login_attempt = Login_Attempt(time.time(), username, client.addr)
+        self.loginAttempts.append(new_login_attempt)
 
         number_of_attempts = len(
-            [
-                i
-                for i in user.loginAttempts
-                if i[0] >= time.time() - self.blacklistThreshold
-            ]
+            self.get_login_attempts(
+                address = client.addr,
+                start_date = time.time() - self.blacklistThreshold,
+            )
         )
 
         if number_of_attempts > self.blacklistLimit:
@@ -703,14 +794,14 @@ class Host:
 
         return False
 
-    async def __got_msg_length__(self, client, data):
+    async def _got_msg_length(self, client, data):
         if not data[7:].isalnum():
             return
         client._next_message_length = int(data[7:])
         if client._next_message_length < self.default_buffer_limit:
             client.reader._limit = client._next_message_length
 
-    async def __got_encData_info__(self, client, data):
+    async def _got_encData_info(self, client, data):
         self.log(
             "{} set encryption to {}".format(
                 client.currentUser.username, data.split(":")[1]
@@ -726,14 +817,14 @@ class Host:
             data = data.decode()
 
         if data.startswith("msgLen=") and len(data) > 7:
-            await self.__got_msg_length__(client, data)
+            await self._got_msg_length(client, data)
         elif data.startswith("LOGIN:") and "|" in data:
             if len(data.split("|")) == 2:
                 data = data[6:]
                 username, password = data.split("|")
 
                 if username in self.users:
-                    success = await self.__got_login_info__(client, username, password)
+                    success = await self._got_login_info(client, username, password)
 
                     if not success:
                         client.disconnect("Failed login")
@@ -744,7 +835,7 @@ class Host:
                     )
                     client.sendRaw(b"login failed")
         elif data.startswith("encData:"):
-            await self.__got_encData_info__(client, data)
+            await self._got_encData_info(client, data)
         elif data == "logout":
             if client.verifiedUser and client.currentUser:
                 client.currentUser.logout(client)
@@ -754,7 +845,7 @@ class Host:
                     )
                 )
 
-    async def __process_data__(self, client, data):
+    async def _process_data(self, client, data):
         if client.verifiedUser and client._usr_enc:
             data = client.currentUser.decryptData(data)
         if data:
@@ -768,13 +859,13 @@ class Host:
                 else:
                     self.gotData(client, data, metaData)
 
-    async def __setup_new_client__(self, reader, writer):
+    async def _setup_new_client(self, reader, writer):
         addr, port = writer.get_extra_info("peername")
         client = Connection(addr, port, reader, writer, self)
         self.clients.append(client)
 
         if self._enable_buffer_monitor:
-            Thread(target=self.__buffer_monitor__, args=[client, reader]).start()
+            Thread(target=self._buffer_monitor, args=[client, reader]).start()
 
         self.log("New Connection: {}:{}".format(client.addr, client.port), "green")
 
@@ -786,9 +877,9 @@ class Host:
 
         return client
 
-    async def __handle_client__(self, reader, writer):
+    async def _handle_client(self, reader, writer):
 
-        client = await self.__setup_new_client__(reader, writer)
+        client = await self._setup_new_client(reader, writer)
 
         if not client.writer.is_closing():
             if self.loginRequired and not client.verifiedUser:
@@ -810,7 +901,7 @@ class Host:
             if not data:
                 break
 
-            await self.__process_data__(client, data)
+            await self._process_data(client, data)
 
         self.log("Lost Connection: {}:{}".format(client.addr, client.port))
         self.clients.remove(client)
@@ -853,7 +944,7 @@ class Host:
                 self.log("SSL certificate loaded", "green")
 
                 server = await asyncio.start_server(
-                    self.__handle_client__,
+                    self._handle_client,
                     self.addr,
                     self.port,
                     ssl=ssl_context,
@@ -865,7 +956,7 @@ class Host:
                 return
         else:
             server = await asyncio.start_server(
-                self.__handle_client__, self.addr, self.port, limit=buffer_limit
+                self._handle_client, self.addr, self.port, limit=buffer_limit
             )
 
         if server:
@@ -881,7 +972,7 @@ class Host:
 class Client:
     sepChar = b"\n\t_SEPARATOR_\t\n"
 
-    def __init__(self, multithreading=False, pickleData=False):
+    def __init__(self, multithreading=False):
         self.connected = False
         self.reader = None
         self.writer = None
@@ -896,7 +987,6 @@ class Client:
         self._next_message_length = 0
         self.default_buffer_limit = 644245094400
         self._enable_buffer_monitor = True
-        self.pickleData = pickleData
 
         self.downloading = False
 
@@ -915,7 +1005,7 @@ class Client:
             self.sendRaw(sData)
             asyncio.run_coroutine_threadsafe(SET_USER_ENCD(self, newValue), self.loop)
 
-    def __start_loop__(self, loop, task, finishFunc):
+    def _start_loop(self, loop, task, finishFunc):
         asyncio.set_event_loop(loop)
         loop.run_until_complete(asyncio.ensure_future(task()))
         if finishFunc:
@@ -923,7 +1013,7 @@ class Client:
 
     def newLoop(self, task, finishFunc=None):
         new_loop = asyncio.new_event_loop()
-        t = Thread(target=self.__start_loop__, args=(new_loop, task, finishFunc))
+        t = Thread(target=self._start_loop, args=(new_loop, task, finishFunc))
         t.start()
 
     def encryptData(self, data):
@@ -966,7 +1056,7 @@ class Client:
                 #       <current buffer length>, <target buffer length>
         return 0
 
-    def __buffer_monitor__(self, reader):
+    def _buffer_monitor(self, reader):
         self.downloading = False
         while self.connected and not self.writer.is_closing():
             if (
@@ -997,14 +1087,14 @@ class Client:
             raise e
         return data.rstrip(self.sepChar)
 
-    async def __got_msg_length__(self, data):
+    async def _got_msg_length(self, data):
         if not data[7:].isalnum():
             return
         self._next_message_length = int(data[7:])
         if self._next_message_length < self.default_buffer_limit:
             self.reader._limit = self._next_message_length
 
-    async def __login_accepted__(self):
+    async def _login_accepted(self):
         self.verifiedUser = True
         if self.multithreading:
             Thread(target=self.loggedIn).start()
@@ -1024,11 +1114,11 @@ class Client:
             data = data.decode()
 
         if data.startswith("msgLen=") and len(data) > 7:
-            await self.__got_msg_length__(data)
+            await self._got_msg_length(data)
         elif data == "login required":
             await self.send_login_info()
         elif data == "login accepted":
-            await self.__login_accepted__()
+            await self._login_accepted()
         elif data == "login failed":
             self._login_failed = True
         elif data == "disconnect":
@@ -1037,7 +1127,7 @@ class Client:
     async def logout(self):
         self.sendRaw(b"logout")
 
-    async def __process_data__(self, data):
+    async def _process_data(self, data):
         if self.login[1] and self.verifiedUser and self._usr_enc:
             data = self.decryptData(data)
         if data:
@@ -1050,9 +1140,9 @@ class Client:
                 else:
                     self.gotData(self, data, metaData)
 
-    async def __handle_host__(self):
+    async def _handle_host(self):
         if self._enable_buffer_monitor:
-            Thread(target=self.__buffer_monitor__, args=[self.reader]).start()
+            Thread(target=self._buffer_monitor, args=[self.reader]).start()
 
         if self.multithreading:
             Thread(target=self.madeConnection).start()
@@ -1065,11 +1155,11 @@ class Client:
                 self.connected = False
                 break
 
-            await self.__process_data__(data)
+            await self._process_data(data)
         self.connected = False
         return self.lostConnection
 
-    async def __handle_self__(self):
+    async def _handle_self(self):
         while self.connected:
             await asyncio.sleep(0.2)
         if not self.connected and self.reader:
@@ -1107,9 +1197,9 @@ class Client:
             self.connection_updated = time.time()
             self.loop = asyncio.get_running_loop()
 
-            future = asyncio.run_coroutine_threadsafe(self.__handle_self__(), self.loop)
+            future = asyncio.run_coroutine_threadsafe(self._handle_self(), self.loop)
 
-            result = self.loop.call_soon_threadsafe(await self.__handle_host__())
+            result = self.loop.call_soon_threadsafe(await self._handle_host())
         except Exception as e:
             print("Error with connection")
             self.connected = False
@@ -1118,7 +1208,7 @@ class Client:
         self.connected = False
         self.connection_updated = time.time()
 
-    async def send_data(self, data, metaData=None):
+    async def _send_data(self, data, metaData=None):
         if not self.connected:
             print("ERROR: Event loop not connected. Unable to send data")
             return None
@@ -1137,7 +1227,7 @@ class Client:
         if self.loop:
             try:
                 asyncio.run_coroutine_threadsafe(
-                    self.send_data(data, metaData=metaData), self.loop
+                    self._send_data(data, metaData=metaData), self.loop
                 )
             except Exception as e:
                 print("Error sending data")
